const _ = require('lodash'),
	CrispCache = require('crisp-cache'),
	cacheLog = require('debug')('cache'),
	MissingResourceError = require('../Error/MissingResourceError'),
	co = require('../Util/co-async'),
	http = require('http'),
	winston = require('winston'),
	ElasticsearchTransport = require('winston-elasticsearch'),
	fs = require('fs-extra'),
	path = require('path'),
	WindowBuffer = require('../Util/WindowBuffer'),
	wrapAsync = require('../Util/wrapAsync');

// Make sure env is loaded
require('./env');

module.exports = {
	config: Module('./config'),
	cacheEnabled: sm => sm.get('config.cache.enabled', true),
	env: sm => require('./env'),
	// Aliases for combined layer sets
	// eg radsat=radar,sat
	layerGroupAliases: sm => require('../parseAliases')('LAYER_GROUP_ALIAS_', sm.get('env')),
	// Aliases for single layers
	// eg clouds=satellite
	layerAliases: sm => require('../parseAliases')('LAYER_ALIAS_', sm.get('env')),

	apiServerRouter: sm => require('../Router/ApiServerRouter')(),
	tileServerRouter: sm => require('../Router/TileServerRouter')(),

	requestOptionsMiddleware: sm => require('../RequestOptions/RequestOptionsMiddleware')(),

	requestOptionsBlueGreenTransformer: sm => require('../RequestOptions/BlueGreenTransformer')(
		sm.get('config').blueGreenTransformer
	),
	/**
	 * Logging
	 */
	loggerBootstrap: sm => {
		const loggerBootstrap = require('logger-bootstrap')({
			appName: 'map-api',
			envPrefix: 'MAP_API',
			requestLogMeta: (req, res) => ({
				layer: req.opts.layer
			}),
			errorResponseLogMeta: (err, req, res) => ({
				layer: req.opts && req.opts.layer
			})
		});

		// Remove Loggly transport from requestLog
		// (It's too much data, and it blows through our limits)
		if (loggerBootstrap.requestLogger.transports.loggly) {
			loggerBootstrap.requestLogger.remove('loggly');
		}

		return loggerBootstrap;
	},
	logger: sm => {
		const logger = sm.get('loggerBootstrap').logger;

		// Add ElasticSearch transport
		if (sm.get('config.elasticSearch.enabled')) {
			logger.add(sm.get('ElasticsearchTransport')({
				level: 'warn',
				indexPrefix: 'map-api-app-logs'
			}), null, true);
		}

		return logger;
	},
	requestLogMiddleware: sm => sm.get('loggerBootstrap').requestLogMiddleware,
	errorResponseLogMiddleware: sm => sm.get('loggerBootstrap').errorResponseLogMiddleware,
	ElasticsearchTransport: sm => options => {
		const envLog = {
			podId: process.env.HOSTNAME || 'unknown',
			type: process.env.DEPLOYMENT_SERVER_TYPE || 'unknown',
			name: process.env.DEPLOYMENT_NAME || 'unknown',
			namespace: process.env.DEPLOYMENT_NAMESPACE || 'unknown',
			image: process.env.DEPLOYMENT_IMAGE || 'unknown',
			version: process.env.DEPLOYMENT_VERSION || 'unknown'
		};
		return new ElasticsearchTransport(Object.assign({
			level: 'info',
			// See https://github.com/elasticsearch/elasticsearch-js/issues/33
			clientOpts: Object.assign({}, sm.get('config.elasticSearch.client')),
			// Add meta data about environment to all logs
			transformer: data => {
				data = require('winston-elasticsearch/transformer')(data);
				data.fields || (data.fields = {});
				data.fields.env = envLog;

				return data;
			}
		}, options));
	},

	httpMetricsLogger: sm => new winston.Logger({
		transports: sm.get('config.elasticSearch.enabled') ? [
			sm.get('ElasticsearchTransport')({
				indexPrefix: 'map-api-http-metrics',
			})
		]: []
	}),
	cacheMetricsLogger: sm => new winston.Logger({
		transports: sm.get('config.elasticSearch.enabled') ? [
			sm.get('ElasticsearchTransport')({
				indexPrefix: 'map-api-cache-metrics',
			})
		] : []
	}),
	requestOptionsLogger: sm => new winston.Logger({
		transports: sm.get('config.elasticSearch.enabled') ? [
			sm.get('ElasticsearchTransport')({
				indexPrefix: 'map-api-request-options',
			})
		] : []
	}),

	ResponseHeaderLogger: sm => (res, lyrId) => {
		const HeaderLogTransport = require('winstonjs-transport-http-headers');
		return new winston.Logger({
			transports: [
				// Here is our new logger
				new HeaderLogTransport({
					silent: !sm.get('config.headerLog.enabled'),
					setHeader: (field, value) => res.set(field, value),
					headerPrefix: lyrId ? `x-aeris-tile-log-${lyrId}- ` : `x-aeris-api-log-`
				})
			]
		});
	},

	tileServerMetricsBuffer: sm => WindowBuffer(sm.get('config.tileServer.metricsAggregationWindow')),

	apiServerMetricsBuffer: sm => WindowBuffer(sm.get('config.apiServer.metricsAggregationWindow')),
	
	/**
	 * Event bindings to log CrispCache hits/misses.
	 * Usage:
	 * 	new CrispCache({
	 * 		events: crispCacheLogEvents('myComponentName')
	 * 	});
	 */
	crispCacheLogEvents: sm => name => {
		const moment = require('moment');
		const logTime = date => moment(date).format('M/DD h:mm:ssa');
		const logTtl = interval => moment.duration(interval).humanize(true);

		return {
			hit: evt => cacheLog(
				`${name} cache hit: "${evt.key}". Created: ${logTime(evt.entry.created)}. ` +
				`Stale ${logTtl(evt.entry.staleTtl)}. Expires ${logTtl(evt.entry.expiresTtl)}. `
			),
			miss: evt => cacheLog(
				`${name} cache miss: ${evt.key}`
			),
			fetchDone: evt => cacheLog(
				`${name}: adding ${evt.key} to cache. ${JSON.stringify(evt.options)} `
			)
		};
	},
	remoteTileServerManager: function() {
		var RemoteTileServerManager = require('../RemoteTileServer/RemoteTileServerManager');

		return new RemoteTileServerManager();
	},
	// Add this after all your other middleware to catch 404s
	'404Handler': function() {
		var MissingResourceError = require('../Error/MissingResourceError');

		return function (req, res, next) {
			if (!res.headersSent) {
				next(new MissingResourceError(req.originalUrl + ' is not a valid endpoint.'));
			}
			else {
				next();
			}
		}
	},

	// Default error handler
	errorHandler: function(sm) {
		var ErrorHandler = require('../ErrorHandler/ErrorHandler');

		return ErrorHandler(sm.get('config').errors);
	},
	warmCache: sm => require('../Util/TileCacheWarmer')(sm.get('config.cache.warmCache')),
	geocode: function(sm) {
		var aerisApiKeys = sm.get('config').aerisApi;
		var Geocoder = require('../Util/Geocoder');

		const geocode = Geocoder({
			clientId: aerisApiKeys.clientId,
			clientSecret: aerisApiKeys.clientSecret
		});
		const cacheOptions = sm.get('config.cache.geocoder', {});

		return sm.get('cacheEnabled') ?
			CrispCache.wrap(geocode, _.extend(cacheOptions, {
				createKey: (place) => place,
				parseKey: (key) => [key],
				events: sm.get('crispCacheLogEvents')('geocode'),
				getOptions: (placeJson) => ({ size: JSON.stringify(placeJson).length })
			})) :
			geocode;
	},
	authService: sm => {
		const AerisAuthService = require('../Auth/AerisAuthService');
		const CachedAerisAuthService = require('../Auth/CachedAerisAuthService');

		const opts = {
			url: sm.get('config.aerisApi.endpoints.permissions'),
			cache: sm.get('config.cache.aerisAuth')
		};

		return sm.get('cacheEnabled') ?
			new CachedAerisAuthService(opts) : new AerisAuthService(opts);
	},
	authMiddleware: sm => require('../Auth/AuthMiddleware')({
		authService: sm.get('authService')
	}),
	dataTimeResolver: sm => {
		const dataTimeResolver = require('../RemoteTileServer/dataTimeResolver');

		if (!sm.get('cacheEnabled')) { return dataTimeResolver; }

		// wrap in cache
		// These methods can't cache for long,
		// but they're a little CPU intensive.
		return _.extend({}, dataTimeResolver, {
			resolveStepOffset: CrispCache.wrap(dataTimeResolver.resolveStepOffset, {
				createKey: (offset, layers) => [offset, layers.join('/')].join('__'),
				parseKey: (key) => {
					var offset, layers;
					const parts = key.split('__');
					return [
						offset = parseInt(parts[0]),
						layers = parts[1].split('/')
					];
				},
				defaultExpiresTtl: 1000 * 10,
				events: sm.get('crispCacheLogEvents')('dataTimeResolver.resolveStepOffset')
			}),
			getClosestDataTime: CrispCache.wrap(dataTimeResolver.getClosestDataTime, {
				createKey: (layerId, validTime) => [layerId, validTime.getTime()].join('__'),
				parseKey: (key) => {
					var parts = key.split('__');
					return [parts[0], new Date(parseInt(parts[1]))];
				},
				defaultExpiresTtl: 1000 * 10,
				events: sm.get('crispCacheLogEvents')('dataTimeResolver.getClosestDataTime')
			})
		});
	},

	// Our remote manager of our projects
	mapProjectApi: sm => {
		const MapProjectApi = require('../Project/MapProjectApi');
		const CachedMapProjectApi = require('../Project/CachedMapProjectApi');

		const opts = {
			host: sm.get('config.mapProjectApi.host'),
			port: sm.get('config.mapProjectApi.port'),
			accessToken: sm.get('config.mapProjectApi.accessToken'),
			cache: sm.get('config.cache.mapProjectApi')
		};
		var projRepo = sm.get('cacheEnabled') ?
			new CachedMapProjectApi(opts) : new MapProjectApi(opts);

		// Log ProjectApi requests
		const logger = sm.get('logger');
		_.extend(projRepo, {
			_request: wrapAsync(projRepo._request.bind(projRepo), {
				before: opts => logger.verbose(`[ProjectApi] Sending request: ${opts.endpoint}?${JSON.stringify(opts.query || {})}`),
				onError: err => logger.warn(`[ProjectApi] Request failed: ${err.message}`, {
					// MapProjectApi._request includes some extra err metadata, for logging
					url: err.url,
					body: err.body,
					statusCode: err.statusCode
				})
			})
		});

		return projRepo;
	},

	TileSourceRepo: sm => {
		// Reuse created tileSourceRepos
		var repos = {};

		return layerId => repos[layerId]
		|| (repos[layerId] = require('../TileSource/TileSourceRepo')({
			layerId: layerId,
			mapProjectApi: sm.get('mapProjectApi'),
			getTileTtl: params => Math.max(params.expires * 1000 - Date.now(), 0)
		}));
	},

	tileServer: sm => require('../Server/TileServer')(),
	tilelive: sm => {
		const tilelive = require('tilelive');

		// Register tilelive modules
		sm.get('config.tileliveModules')
			// Load modules
			.map(moduleName => require(moduleName))
			// Register modules
			.forEach(tileliveModule => _.isFunction(tileliveModule.registerProtocols) ?
				tileliveModule.registerProtocols(tilelive) : tileliveModule(tilelive)
			);

		return sm.get('cacheEnabled') ?
			// Create cached tilelive wrapper
			require('../TileSource/CachedTileLive')(tilelive, {
				cache: sm.get('config').cache.tileSources,
				tileCache: sm.get('config').cache.tiles
			}) :
			tilelive;
	},

	apiServerHttpCache: sm => {
		const CrispHttpCache = require('crisp-http-cache');
		const Location = require('../RequestOptions/Type/Location');
		const RequestOptions = require('../RequestOptions/RequestOptions');

		return new CrispHttpCache({
			enabled: sm.get('cacheEnabled'),
			getKey: (req, res, cb) => {
				// Do not include client in cache keys,
				// so we can cache across clients
				const clientUnawareUrl = Object.assign({}, req.opts, {
					clientId: 'CACHE_CLIENT_ID',
					clientSecret: 'CACHE_CLIENT_SECRET'
				});
				cb(null, RequestOptions.toUrl(clientUnawareUrl));
			},
			// don't change my headers, please
			transformHeaders: () => {},
			cacheOptions: {
				emitEvents: true,
				maxSize: sm.get('config.cache.apiServerHttp.maxSize', 0)
			}
		});
	},
	tileServerHttpCache: sm => {
		const CrispHttpCache = require('crisp-http-cache');

		var cache = new CrispHttpCache({
			enabled: sm.get('cacheEnabled'),
			cacheOptions: {
				emitEvents: true,
				maxSize: sm.get('config.cache.tileServerHttp.maxSize', 0)
			},
			// don't change my headers, please
			transformHeaders: () => {}
		});

		if(sm.get('config').cache.debug) {
			cache.backend.on('hit', debugCacheUsage(cache, 'Tile'));
			cache.backend.on('miss', debugCacheUsage(cache, 'Tile'));
		}
		return cache;
	},
	tileServerTimesHttpCache: sm => {
		const CrispHttpCache = require('crisp-http-cache');

		return new CrispHttpCache({
			enabled: sm.get('cacheEnabled'),
			cacheOptions: {
				maxSize: sm.get('config.cache.tileServerHttp.timesMaxSize', 0)
			},
			// don't change my headers, please
			transformHeaders: () => {}
		});
	},

	remoteTileServerCaches: sm => ({}),
	RemoteTileServer: sm => sm.get('cacheEnabled') ?
		function(name, dns, port) {
			const instance = require('../RemoteTileServer/CachedRemoteTileServer')(name, dns, port);

			// Save instance, so we can keep cache metrics
			const caches = sm.get('remoteTileServerCaches');
			caches[`remoteTileServer.${name}.requestTimes`] = instance.requestTimes._cache;

			return instance
		} :
		require('../RemoteTileServer/RemoteTileServer'),
	combineImages: Module('../Util/combineImages'),
	mapnik: Module('mapnik'),
	blend: Module('blend'),
	sharp: Module('sharp'),
	abaculus: () => {
		const abaculus = require('abaculus');
		return function(opts, cb) {
			try {
				abaculus(opts, cb)
			}
			catch (err) {
				cb(err);
			}
		}
	},

	// Takes a 'protocol' prefixed path and resolves it to an absolute path
	pathResolver: sm => (projPath) => {
		const path = require('path'),
			rootDir = sm.get('config').projectRootDir;
		
		const localProjPath = path.normalize(projPath.replace(/^nfs:\/\/map-projects/, rootDir));

		// We are sure we didn't ../../ back too far (this is essentially chroot-ing us)
		if (!localProjPath.startsWith(rootDir)) {
			throw new Error(`Failed to resolve project path: path resides outside of root directory (${projPath})`);
		}
		
		return localProjPath;
	},

	// All of the CrispCache instances used by the TileServer
	// (so we can grab metrics from them)
	tileServerCaches: sm => sm.get('cacheEnabled') ? {
		'http.image': sm.get('tileServerHttpCache').backend,
		'http.times': sm.get('tileServerTimesHttpCache').backend,
		'tilelive.tileSource': sm.get('tilelive').tileSourceCache,
		'tilelive.tile': sm.get('tilelive').tileCache,
		'mapProjectApi.projects': sm.get('mapProjectApi').getProjects._cache,
		'mapProjectApi.dataSet': sm.get('mapProjectApi').getDataSet._cache,
		'mapProjectApi.layer': sm.get('mapProjectApi').getLayer._cache,
	} : {},
	apiServerCaches: sm => sm.get('cacheEnabled') ? Object.assign(
		{
			'dataTimeResolver.stepOffset': sm.get('dataTimeResolver').resolveStepOffset._cache,
			'dataTimeResolver.closestDataTime': sm.get('dataTimeResolver').getClosestDataTime._cache,
			geocode: sm.get('geocode')._cache,
			authPermissions: sm.get('authService').fetchPermissions._cache,
			http: sm.get('apiServerHttpCache').backend
		},
		sm.get('remoteTileServerCaches')
	) : {}
};

function Module(name) {
	return () => require(name);
}



function toEnvCase(str) {
	return str.
		match(/[a-z|0-9]+/g).
		map(function(word) {
			return word.toUpperCase();
		}).
		join('_');
}

function debugCacheUsage(cache, name) {
	return function() {
		var usage = cache.getUsage();
		console.log(`${name} Cache Usage: ${Math.round(usage.size/1024/1024*100)/100}MB of ${Math.round(usage.maxSize/1024/1024*100)/100}MB ${Math.round(usage.size/usage.maxSize * 10000) / 100}%`)
	}
}